@@ -43,6 +43,7 @@ from typing import MutableSequence
 from typing import Optional
 from typing import overload
 from typing import Sequence
+from typing import Sized
 from typing import Tuple
 from typing import TypeVar
 from typing import Union
@@ -51,19 +52,69 @@ from typing import Union
 _T_co = TypeVar("_T_co", covariant=True)
 
 
+def _positive(index: int, size: int) -> int:
+    """Convert a negative index to a non-negative integer.
+
+    The index is interpreted relative to the position after the last item. If
+    the index is smaller than ``-size``, IndexError is raised.
+    """
+    assert index < 0  # noqa: S101
+    index += size
+
+    if index < 0:
+        raise IndexError("lazysequence index out of range")
+
+    return index
+
+
+def _positivestart(start: int, size: int) -> int:
+    """Convert a negative start index to a non-negative integer.
+
+    The index is interpreted relative to the position after the last item. If
+    the index is smaller than ``-size``, zero is returned.
+    """
+    assert start < 0  # noqa: S101
+    return max(0, start + size)
+
+
+def _positivestop(stop: int, size: int, step: int) -> Optional[int]:
+    """Convert a negative stop index to a non-negative integer.
+
+    The index is interpreted relative to the position after the last item. If
+    the index is smaller than ``-size``, the return value depends on ``step``.
+    If ``step`` is positive, zero is returned. If ``step`` is negative, None is
+    returned. Returning None for reverse slices ensures that an expression such
+    as ``s[:-1000:-1]`` includes the first element of ``s``.
+    """
+    assert stop < 0  # noqa: S101
+
+    stop += size
+    if stop >= 0:
+        return stop
+
+    return 0 if step > 0 else None
+
+
 @dataclass(frozen=True)
 class _slice:  # noqa: N801
+    """Slice representation.
+
+    Attributes:
+        start: The position of the first item, either None or a non-negative
+            integer.
+        stop: The position after the last item, either None or a non-negative
+            integer.
+        step: The distance between successive items, any integer except zero.
+    """
+
     start: Optional[int]
     stop: Optional[int]
     step: int
 
-    @classmethod
-    def fromslice(cls, aslice: slice) -> _slice:
-        return cls(aslice.start, aslice.stop, aslice.step)
-
     def __init__(
         self, start: Optional[int], stop: Optional[int], step: Optional[int]
     ) -> None:
+        """Initialize."""
         if step == 0:
             raise ValueError("slice step cannot be zero")
 
@@ -74,46 +125,48 @@ class _slice:  # noqa: N801
         object.__setattr__(self, "stop", stop)
         object.__setattr__(self, "step", step)
 
-    def asslice(self) -> slice:
-        return slice(*self.astuple())
+    def astuple(self) -> Tuple[Optional[int], Optional[int], int]:
+        """Return the attributes as a tuple."""
+        return self.start, self.stop, self.step
 
-    def hasnegativebounds(self) -> bool:
-        return any(arg < 0 for arg in (self.start, self.stop) if arg is not None)
+    def apply(self, iterable: Iterable[_T_co], sized: Sized) -> Iterator[_T_co]:
+        """Yield items from the iterable corresponding to this slice."""
+        return islice(iterable, *self.astuple())
 
-    def withpositivebounds(self, size: int) -> _slice:
-        start, stop, step = self.astuple()
+    def length(self, sized: Sized) -> int:
+        """Return the number of items corresponding to this slice."""
+        origin = self
 
-        if start is not None and start < 0:
-            start = max(0, start + size)
+        if origin.step < 0:
+            origin = origin.reverse(sized)
 
-        if stop is not None and stop < 0:
-            stop += size
-            if stop < 0:
-                stop = None if step < 0 else 0
+        start, stop, step = origin.astuple()
+        size = len(sized)
 
-        return _slice(start, stop, step)
+        if stop is not None:
+            size = min(size, stop)
 
-    def length(self, size: int) -> int:
-        if self.stop is not None:
-            size = min(size, self.stop)
-
-        if self.start is not None:
-            size = max(0, size - self.start)
+        if start is not None:
+            size = max(0, size - start)
 
         if size > 0:
             # This is equivalent to `math.ceil(result / step)`, but avoids
             # floating-point operations and importing `math`.
-            size = 1 + (size - 1) // self.step
+            size = 1 + (size - 1) // abs(step)
 
         return size
 
-    def reverse(self, size: int) -> _slice:
+    def reverse(self, sized: Sized) -> _slice:
+        """Return the equivalent slice for a reversed sequence."""
         assert self.step < 0  # noqa: S101
 
         start, stop, step = self.astuple()
+        size = len(sized)
         step = -step
 
-        assert start is not None  # noqa: S101
+        if start is None:
+            start = size - 1
+
         start = (size - 1) - start
         start = max(0, start)
 
@@ -126,13 +179,25 @@ class _slice:  # noqa: N801
 
         return _slice(start, stop, step)
 
-    def astuple(self) -> Tuple[Optional[int], Optional[int], int]:
-        return self.start, self.stop, self.step
+    def resolve(self, index: int, sized: Sized) -> int:
+        """Return the equivalent index on the underlying sequence.
 
-    def apply(self, iterable: Iterable[_T_co]) -> Iterator[_T_co]:
-        return islice(iterable, *self.astuple())
+        In pseudo-code: ``s[slice][index] === s[slice.resolve(index)]``
+        """
+        if index < 0:
+            index = _positive(index, self.length(sized))
 
-    def resolve(self, index: int) -> int:
+        return self._resolve(index, sized, strict=True)
+
+    def _resolve(self, index: int, sized: Sized, *, strict: bool = True) -> int:
+        assert index >= 0  # noqa: S101
+        return (
+            self._resolveforward(index, sized, strict=strict)
+            if self.step > 0
+            else self._resolvebackward(index, sized, strict=strict)
+        )
+
+    def _resolveforward(self, index: int, sized: Sized, *, strict: bool = True) -> int:
         """Resolve index on a forward slice, where start <= stop and step > 0."""
         assert self.step > 0  # noqa: S101
 
@@ -144,44 +209,85 @@ class _slice:  # noqa: N801
         index = start + index * step
 
         if stop is not None and index >= stop:
-            raise IndexError("lazysequence index out of range")
+            if strict:
+                raise IndexError("lazysequence index out of range")
+
+            # Default to the last valid slot.
+            return max(0, stop - 1)
 
         return index
 
-    def resolve_noraise(self, index: int) -> Optional[int]:
-        """Resolve index, defaulting to the last valid slot."""
-        assert self.step > 0  # noqa: S101
-
-        try:
-            return self.resolve(index)
-        except IndexError:
-            return self.stop - 1 if self.stop is not None else None
-
-    def rresolve(self, index: int, size: int) -> int:
+    def _resolvebackward(self, index: int, sized: Sized, *, strict: bool = True) -> int:
         """Resolve index on a backward slice, where start >= stop and step < 0."""
         assert self.step < 0  # noqa: S101
 
+        size = len(sized)
         start, stop, step = self.astuple()
 
-        assert start is not None  # noqa: S101
+        if start is None:
+            start = size - 1
 
         start = min(start, size - 1)
         index = start + index * step
 
         if index < 0 or stop is not None and index <= stop:
-            raise IndexError("lazysequence index out of range")
+            if strict:
+                raise IndexError("lazysequence index out of range")
+
+            # Default to the first valid slot.
+            return stop + 1 if stop is not None else 0
 
         return index
 
-    def rresolve_noraise(self, index: int, size: int) -> int:
-        """Resolve index backwards, defaulting to the first valid slot."""
-        try:
-            return self.rresolve(index, size)
-        except IndexError:
-            return self.stop + 1 if self.stop is not None else 0
+    def resolveslice(self, slice: _slice, sized: Sized) -> _slice:
+        """Return the equivalent slice on the underlying sequence.
+
+        In pseudo-code: ``s[slice0][slice1] === s[slice0.resolve(slice1)]``
+        """
+        start, stop, step = slice.astuple()
+
+        start = self._resolvestart(start, step, sized)
+        stop = self._resolvestop(stop, step, sized)
+        step *= self.step
+
+        return _slice(start, stop, step)
+
+    def _resolvestart(
+        self, start: Optional[int], step: int, sized: Sized
+    ) -> Optional[int]:
+        if start is not None and start < 0:
+            start = _positivestart(start, self.length(sized))
+
+        if start is not None:
+            return self._resolve(start, sized, strict=False)
+
+        if step > 0:
+            return self.start
+
+        if self.stop is None:
+            return None
+
+        return self.stop + 1 if self.step < 0 else self.stop - 1
+
+    def _resolvestop(
+        self, stop: Optional[int], step: int, sized: Sized
+    ) -> Optional[int]:
+        if stop is not None and stop < 0:
+            stop = _positivestop(stop, self.length(sized), step)
+
+        if stop is not None:
+            return self._resolve(stop, sized, strict=False)
+
+        if step > 0:
+            return self.stop
+
+        if self.start is None:
+            return None
+
+        return self.start + 1 if self.step < 0 else self.start - 1
 
 
-_defaultslice = slice(None)
+_defaultslice = _slice(None, None, None)
 
 
 class lazysequence(Sequence[_T_co]):  # noqa: N801
@@ -200,41 +306,44 @@ class lazysequence(Sequence[_T_co]):  # noqa: N801
         *,
         storage: Callable[[], MutableSequence[_T_co]] = deque,
         _cache: Optional[MutableSequence[_T_co]] = None,
-        _indices: slice = _defaultslice,
+        _slice: _slice = _defaultslice,
     ) -> None:
         """Initialize."""
+        parent = self
+
+        class _Total(Sized):
+            def __len__(self) -> int:
+                parent._fill()
+                return len(parent._cache)
+
         self._iter = iter(iterable)
         self._cache = storage() if _cache is None else _cache
-
-        slice = _slice.fromslice(_indices)
-
-        if slice.hasnegativebounds():
-            self._fill()
-            self._slice = slice.withpositivebounds(self._cachesize)
-        else:
-            self._slice = slice
+        self._slice = _slice
+        self._total = _Total()
 
     def _consume(self) -> Iterator[_T_co]:
+        """Yield from the iterator, adding each item to the cache."""
         for item in self._iter:
             self._cache.append(item)
             yield item
 
     def _fill(self) -> None:
+        """Add all items from the iterator to the cache."""
         self._cache.extend(self._iter)
 
-    @property
-    def _cachesize(self) -> int:
-        return len(self._cache)
-
     def _iterate(self, iterator: Iterator[_T_co]) -> Iterator[_T_co]:
+        """Iterate over the items in the sequence."""
         slice = self._slice
 
-        if slice.step < 0:
-            self._fill()
-            slice = slice.reverse(self._cachesize)
-            return slice.apply(reversed(self._cache))
+        if slice.step > 0:
+            iterator = chain(self._cache, iterator)
+        else:
+            slice = slice.reverse(self._total)
 
-        return slice.apply(chain(self._cache, iterator))
+            self._fill()
+            iterator = reversed(self._cache)
+
+        return slice.apply(iterator, self._total)
 
     def __iter__(self) -> Iterator[_T_co]:
         """Iterate over the items in the sequence."""
@@ -259,34 +368,18 @@ class lazysequence(Sequence[_T_co]):  # noqa: N801
 
     def __len__(self) -> int:
         """Return the number of items in the sequence."""
-        slice = self._slice
-
-        self._fill()
-
-        if slice.step < 0:
-            slice = slice.reverse(self._cachesize)
-
-        return slice.length(self._cachesize)
+        return self._slice.length(self._total)
 
     def _getitem(self, index: int) -> _T_co:
-        if index < 0:
-            index += len(self)
-
-        if index < 0:
-            raise IndexError("lazysequence index out of range")
-
-        if self._slice.step >= 0:
-            index = self._slice.resolve(index)
-        else:
-            self._fill()
-            index = self._slice.rresolve(index, self._cachesize)
+        """Return the item at the given index."""
+        index = self._slice.resolve(index, self._total)
 
         try:
             return self._cache[index]
         except IndexError:
             pass
 
-        index -= self._cachesize
+        index -= len(self._cache)
 
         try:
             return next(islice(self._consume(), index, None))
@@ -294,53 +387,11 @@ class lazysequence(Sequence[_T_co]):  # noqa: N801
             raise IndexError("lazysequence index out of range") from None
 
     def _getslice(self, indices: slice) -> lazysequence[_T_co]:  # noqa: C901
-        def resolve(index: int) -> Optional[int]:
-            if self._slice.step > 0:
-                return self._slice.resolve_noraise(index)
+        """Return a slice of the sequence."""
+        slice = _slice(indices.start, indices.stop, indices.step)
+        slice = self._slice.resolveslice(slice, self._total)
 
-            self._fill()
-            return self._slice.rresolve_noraise(index, self._cachesize)
-
-        def resolve_start(start: Optional[int], step: int) -> Optional[int]:
-            assert start is None or start >= 0  # noqa: S101
-
-            if start is None:
-                start = 0 if step > 0 else len(self) - 1
-
-            return resolve(start)
-
-        def resolve_stop(stop: Optional[int], step: int) -> Optional[int]:
-            assert stop is None or stop >= 0  # noqa: S101
-
-            if stop is not None:
-                return resolve(stop)
-
-            if step > 0:
-                return self._slice.stop
-
-            if self._slice.start is None:
-                return None
-
-            if self._slice.step < 0:
-                return self._slice.start + 1
-
-            return self._slice.start - 1
-
-        def resolve_slice(aslice: _slice) -> _slice:
-            if aslice.hasnegativebounds():
-                aslice = aslice.withpositivebounds(len(self))
-
-            start, stop, step = aslice.astuple()
-
-            start = resolve_start(start, step)
-            stop = resolve_stop(stop, step)
-            step *= self._slice.step
-
-            return _slice(start, stop, step)
-
-        theslice = resolve_slice(_slice.fromslice(indices))
-
-        return lazysequence(self._iter, _cache=self._cache, _indices=theslice.asslice())
+        return lazysequence(self._iter, _cache=self._cache, _slice=slice)
 
     @overload
     def __getitem__(self, index: int) -> _T_co:
@@ -353,7 +404,7 @@ class lazysequence(Sequence[_T_co]):  # noqa: N801
     def __getitem__(
         self, index: Union[int, slice]
     ) -> Union[_T_co, lazysequence[_T_co]]:
-        """Return the item at the given index."""
+        """Return the item at the given index, or a slice of the sequence."""
         return (
             self._getslice(index) if isinstance(index, slice) else self._getitem(index)
         )
